import functools
import asyncio
from .baseAdvice import BaseAdvice

class Aspect:
    def __init__(self):
        self.advice = None

    def set_advice(self, advice: BaseAdvice):
        if not isinstance(advice, BaseAdvice):
            raise TypeError("advice must inherit from BaseAdvice.")
        self.advice = advice

    async def _call_advice_async(self, method, *args, **kwargs):
        if method is None:
            return None
        if asyncio.iscoroutinefunction(method):
            return await method(*args, **kwargs)
        else:
            result = method(*args, **kwargs)
            if asyncio.iscoroutine(result):
                return await result
            return result

    def _call_advice_sync(self, method, *args, **kwargs):
        if method is None:
            return None
        if asyncio.iscoroutinefunction(method):
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                return asyncio.run(method(*args, **kwargs))
            else:
                return loop.run_until_complete(method(*args, **kwargs))
        else:
            result = method(*args, **kwargs)
            if asyncio.iscoroutine(result):
                try:
                    loop = asyncio.get_running_loop()
                except RuntimeError:
                    return asyncio.run(result)
                else:
                    return loop.run_until_complete(result)
            return result

    async def execute_advice_async(self, func, args, kwargs):
        try:
            if self.advice:
                await self._call_advice_async(self.advice.before, func, *args, **kwargs)

            if self.advice and hasattr(self.advice, "around"):
                result = await self._call_advice_async(self.advice.around, func, *args, **kwargs)
            else:
                result = await func(*args, **kwargs)

            if self.advice:
                await self._call_advice_async(self.advice.after, func, result, *args, **kwargs)
            return result
        except Exception as e:
            if self.advice:
                await self._call_advice_async(self.advice.on_exception, func, e, *args, **kwargs)
            raise e  # 예외를 다시 발생시켜 메인 except로 전달

    def execute_advice_sync(self, func, args, kwargs):
        try:
            if self.advice:
                self._call_advice_sync(self.advice.before, func, *args, **kwargs)

            if self.advice and hasattr(self.advice, "around"):
                result = self._call_advice_sync(self.advice.around, func, *args, **kwargs)
            else:
                result = func(*args, **kwargs)

            if self.advice:
                after_result = self._call_advice_sync(self.advice.after, func, result, *args, **kwargs)
                if asyncio.iscoroutine(after_result):
                    try:
                        loop = asyncio.get_running_loop()
                    except RuntimeError:
                        asyncio.run(after_result)
                    else:
                        loop.create_task(after_result)
            return result
        except Exception as e:
            if self.advice:
                self._call_advice_sync(self.advice.on_exception, func, e, *args, **kwargs)
            raise e  # 예외를 다시 발생시켜 메인 except로 전달

    def apply(self, func):
        if asyncio.iscoroutinefunction(func):
            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                return await self.execute_advice_async(func, args, kwargs)
            return async_wrapper
        else:
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return self.execute_advice_sync(func, args, kwargs)
            return wrapper


from abc import ABC, abstractmethod


class BaseAdvice(ABC):
    """
    AOP Advice Interface
    """

    @abstractmethod
    def before(self, func, *args, **kwargs):
        pass

    @abstractmethod
    def after(self, func, result, *args, **kwargs):
        pass

    @abstractmethod
    def around(self, func, *args, **kwargs):
        pass

    @abstractmethod
    def on_exception(self, func, exception, *args, **kwargs):
        pass


import functools
import time

def before(advice):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            advice(func, *args, **kwargs)
            return func(*args, **kwargs)
        return wrapper
    return decorator

def after(advice):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            advice(func, result, *args, **kwargs)
            return result
        return wrapper
    return decorator

def around(advice):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return advice(func, *args, **kwargs)
        return wrapper
    return decorator

def on_exception(advice):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                advice(func, e, *args, **kwargs)
                raise e
        return wrapper
    return decorator


from setuptools import setup, find_packages

setup(
    name="aophelper",
    version="0.6.0",  # 버전
    author="mathbook3948",  # 작성자
    author_email="yuno.jung.07@gmail.com",  # 이메일
    description="AOP (Aspect-Oriented Programming) helper for Python",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/mathbook3948/AOPHelper",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[],
)